package org.xbib.interlibrary.http.server.avail;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.function.Predicate;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.interlibrary.action.avail.AvailAction;
import org.xbib.interlibrary.action.avail.AvailRequest;
import org.xbib.interlibrary.action.avail.AvailRequester;
import org.xbib.interlibrary.action.avail.AvailResponse;
import org.xbib.interlibrary.action.avail.DefaultAvailRequester;
import org.xbib.interlibrary.api.library.Library;
import org.xbib.interlibrary.api.library.LibraryService;
import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.api.library.Mode;
import org.xbib.interlibrary.api.util.Maps;
import org.xbib.interlibrary.common.DefaultBibliographicDescription;
import org.xbib.interlibrary.common.predicate.MaterialPredicate;
import org.xbib.interlibrary.common.predicate.DomainPredicate;
import org.xbib.interlibrary.common.predicate.LastResortPredicate;
import org.xbib.interlibrary.common.predicate.LibraryGroupPredicate;
import org.xbib.interlibrary.common.predicate.ModePredicate;
import org.xbib.interlibrary.common.predicate.OrganizationPredicate;
import org.xbib.interlibrary.http.server.InterlibraryServerApplicationModule;
import org.xbib.net.ParameterException;
import org.xbib.net.http.server.HttpRequest;
import org.xbib.net.http.server.application.Application;
import org.xbib.net.http.server.route.HttpRouterContext;
import org.xbib.net.http.server.session.Session;
import org.xbib.settings.Settings;

public class AvailApplicationModule extends InterlibraryServerApplicationModule {

    private static final Logger logger = Logger.getLogger(AvailApplicationModule.class.getName());

    private final HttpAvailServer httpAvailServer;

    private final AvailModule availModule;

    public AvailApplicationModule(HttpAvailServer httpAvailServer,
                                  Application application,
                                  String name,
                                  Settings settings) {
        super(application, name, settings);
        this.httpAvailServer = httpAvailServer;
        this.availModule = httpAvailServer.getModule(AvailModule.class);
    }

    public AvailResponse availRequest(HttpRequest httpRequest) {
        Objects.requireNonNull(httpRequest);
        try {
            return availModule.avail(httpRequest.getParameter().asMultiMap());
        } catch (ParameterException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        return null;
    }

    public String getFormId(String key, HttpRequest request) {
        return AvailForm.getFormId(key, request.getParameter());
    }

    public Map<String, Object> getAvailForm(String key, HttpRequest request, Session session) {
        TinyMap.Builder<String, Object> builder = AvailForm.getForm(key, request.getParameter(), session);
        Collection<?> currentValue = AvailForm.getList(builder, "source");
        logger.log(Level.INFO, "current value = " + currentValue);
        builder.put("sources", httpAvailServer.getAvailServices(currentValue));
        Map<String, Object> map = builder.build();
        String id = (String) map.get("_id");
        if (id == null) {
            logger.log(Level.WARNING, "no ID found for session update, map = " + map);
            return map;
        }
        session.put(id, map);
        return map;
    }

    @SuppressWarnings("unchecked")
    public Map<String, Object> getAvailResponse(String key, Session session) {
        return (Map<String, Object>) session.get(key);
    }

    public void removeAvailResponse(String key, Session session) {
        session.remove(key);
    }

    @SuppressWarnings("unchecked")
    public void executeAvail(String key, HttpRouterContext context) {
        HttpRequest request = context.getRequest();
        Session session = context.getAttributes().get(Session.class, "session");
        Map<String, Object> map = getAvailForm(key, request, session);
        String requester = (String) map.get("requester");
        if (requester == null) {
            requester = "DE-605";
        }
        Collection<?> services = AvailForm.getList(map, "db");
        String source = (String) map.get("source");
        if (source == null || source.isEmpty()) {
            logger.log(Level.WARNING, "source is empty, nothing to do");
            return;
        }
        String sourceId = (String) map.get("sourceid");
        if (sourceId == null || sourceId.isEmpty()) {
            logger.log(Level.WARNING, "source ID is empty, nothing to do");
            return;
        }
        String yearString = (String) map.get("year");
        Integer year = yearString != null && !yearString.isEmpty() ? Integer.parseInt(yearString) : null;
        DefaultBibliographicDescription.Builder bibliographicDescriptionBuilder = DefaultBibliographicDescription.builder()
                .setSource(source)
                .setSourceId(sourceId)
                .setYear(year);
        Predicate<Library> libraryPredicate = library -> true;
        StringBuilder libraryPredicateText = new StringBuilder();
        Predicate<LibraryService> libraryServicePredicate = libraryService -> true;
        StringBuilder libraryServicePredicateText = new StringBuilder();
        String mode = (String) map.get("mode");
        if (mode != null && !mode.isEmpty()) {
            libraryServicePredicate = libraryServicePredicate.and(new ModePredicate(Mode.fromValue(mode)));
            libraryServicePredicateText.append(" AND MODE(").append(mode).append(")");
        } else {
            // dummy
            mode = "";
        }
        String material = (String) map.get("material");
        if (material != null && !material.isEmpty()) {
            libraryServicePredicate = libraryServicePredicate.and(new MaterialPredicate(Material.fromValue(material)));
            libraryServicePredicateText.append(" AND MATERIAL(").append(material).append(")");
        } else {
            // dummy
            material = "";
        }
        String domain = (String) map.get("domain");
        if (domain != null && !domain.isEmpty()) {
            libraryPredicate = libraryPredicate.and(new DomainPredicate(domain));
            libraryPredicateText.append(" AND DOMAIN(").append(domain).append(")");
        } else {
            domain = "";
        }
        String organization = (String) map.get("organization");
        if (organization != null && !organization.isEmpty()) {
            libraryPredicate = libraryPredicate.and(new OrganizationPredicate(organization));
            libraryPredicateText.append(" AND ORGANIZATION(").append(organization).append(")");
        } else {
            organization = "";
        }
        if (material.contains("online_resource")) {
            if (mode.contains("copy")) {
                Predicate<Library> allowEFL = new LibraryGroupPredicate("EFL").or(new DomainPredicate("NRW").negate());
                Predicate<Library> notblockedEFL = new LibraryGroupPredicate("EFLblock").negate();
                libraryPredicate = libraryPredicate.and(allowEFL.and(notblockedEFL));
                libraryPredicateText.append("(GROUP(EFL) OR NOT(DOMAIN(NRW))) AND NOT(GROUP(EFLblock))");
            } else if (mode.contains("loan")) {
                Predicate<Library> allowZFL = new LibraryGroupPredicate("ZFL").or(new DomainPredicate("NRW").negate());
                Predicate<Library> notblockedZFL = new LibraryGroupPredicate("ZFLblock").negate();
                libraryPredicate = libraryPredicate.and(allowZFL.and(notblockedZFL));
                libraryPredicateText.append("(GROUP(ZFL) OR NOT(DOMAIN(NRW))) AND NOT(GROUP(ZFLblock))");
            }
        } else if (material.contains("text")) {
            if (mode.contains("copy")) {
                Predicate<Library> allowSFL = new LibraryGroupPredicate("SFL").or(new DomainPredicate("NRW").negate());
                Predicate<Library> notblockedSFL = new LibraryGroupPredicate("SFLblock").negate();
                libraryPredicate = libraryPredicate.and(allowSFL.and(notblockedSFL));
                libraryPredicateText.append("(GROUP(SFL) OR NOT(DOMAIN(NRW))) AND NOT(GROUP(SFLblock))");
            } else if (mode.contains("loan")) {
                Predicate<Library> allowZFL =
                        new LibraryGroupPredicate("ZFL").or(new DomainPredicate("NRW").negate());
                Predicate<Library> notblockedZFL = new LibraryGroupPredicate("ZFLblock").negate();
                libraryPredicate = libraryPredicate.and(allowZFL.and(notblockedZFL));
                libraryPredicateText.append("(GROUP(ZFL) OR NOT(DOMAIN(NRW))) AND NOT(GROUP(ZFLblock))");
            }
        } else {
            logger.log(Level.FINE, "material " + material + " undefined, setting no predicate");
        }
        libraryPredicate = libraryPredicate.or(new LastResortPredicate());
        libraryPredicateText.append(" OR LASTRESORT()");
        logger.log(Level.FINE, "library predicate = " + libraryPredicateText);
        logger.log(Level.FINE, "library service predicate = " + libraryServicePredicateText);
        List<String> advantage = (List<String>) map.get("advantage");
        List<String> handicap = (List<String>) map.get("handicap");
        AvailRequester availRequester = new DefaultAvailRequester(requester, null,
                Maps.getBoolean(map, "selfallowed"),
                Maps.getBoolean(map, "localallowed"),
                Maps.getBoolean(map, "domainallowed"));
        AvailRequest.Builder availRequestBuilder = AvailRequest.builder()
                .setServices(services)
                .setRequester(availRequester)
                .setBibliographicDescription(bibliographicDescriptionBuilder.build())
                .setDomainEnabled(true)
                .setLibraryPredicate(libraryPredicate)
                .setLibraryPredicateText(libraryPredicateText.toString())
                .setLibraryServicePredicate(libraryServicePredicate)
                .setLibraryServicePredicateText(libraryServicePredicateText.toString())
                .setAdvantage(advantage)
                .setHandicap(handicap)
                .setFilteredLibrariesEnabled(true) // record filtered libraries
                .setBibliographicDescriptionEnabled(true) // required for open access check
                .setLinkPriorityEnabled(true); // evaluate material in bibliographic description
        AvailResponse availResponse = httpAvailServer.execute(AvailAction.INSTANCE, availRequestBuilder.build());
        // for storing the response, we must use the session, in order to span several requests/response
        session.put("_availform", map);
        session.put("_availresponse", availResponse.asMap());
    }
}
