package org.xbib.interlibrary.http.server.avail;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.http.server.session.Session;
import org.xbib.net.util.RandomUtil;

public class AvailForm {

    private static final Logger logger = Logger.getLogger(AvailForm.class.getName());

    private AvailForm() {
    }

    public static String getFormId(String key, Parameter parameter) {
        try {
            String id = parameter.getAsString(key, Parameter.Domain.FORM);
            if (id == null || id.isEmpty()) {
                id = parameter.getAsString(key, Parameter.Domain.PATH);
                if (id == null || id.isEmpty()) {
                    return null;
                }
            }
            return id;
        } catch (ParameterException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            return null;
        }
    }

    @SuppressWarnings("unchecked")
    public static TinyMap.Builder<String, Object> getForm(String key, Parameter parameter, Session session) {
        TinyMap.Builder<String, Object> builder = TinyMap.builder();
        String id = getFormId(key, parameter);
        if (id == null) {
            logger.log(Level.FINER, "no ID found for " + key + ", using new random ID");
            id = RandomUtil.randomString(16);
        }
        Map<String, Object> map = (Map<String, Object> ) session.get(id);
        builder.put("_id", id);
        builder.put("_key", key);
        Object source = current("source", parameter, map, "HBZ");
        Object sourceid = current("sourceid", parameter, map, "");
        Object year = current("year", parameter, map,"");
        Object mode = current("mode", parameter, map, "copy");
        Object material = current("material", parameter, map,"");
        Object domain = current("domain", parameter, map,"");
        Object selfallowed = currentCheck("selfallowed", parameter, map, List.of("false"));
        Object localallowed = currentCheck("localallowed", parameter, map, List.of("false"));
        Object domainallowed = currentCheck("domainallowed", parameter, map, List.of("true"));
        Object advantage = current("advantage", parameter, map, List.of());
        Object handicap = current("handicap", parameter, map, List.of());
        builder.put("source", source);
        builder.put("sourceid", sourceid);
        builder.put("year", year);
        builder.put("mode", mode);
        builder.put("modes", modes(mode));
        builder.put("material", material);
        builder.put("materials", materials(material));
        builder.put("domain", domain);
        builder.put("domains", domains(domain));
        builder.put("selfallowed", selfallowed);
        builder.put("localallowed", localallowed);
        builder.put("domainallowed", domainallowed);
        builder.put("advantage", advantage);
        builder.put("handicap", handicap);
        return builder;
    }

    private static List<Key> modes(Object current) {
        return List.of(
                Key.of("Leihe, Kopie",  "loan_copy", "loan_copy".equals(current), false),
                Key.of("Leihe",  "loan", "loan".equals(current), false),
                Key.of("Kopie", "copy", "copy".equals(current), false),
                Key.of("elektronische Kopie",  "electronic_copy", "electronic_copy".equals(current), false),
                Key.of("keine Fernleihe erlaubt", "none", "none".equals(current), false),
                Key.of("nicht definiert", "", "".equals(current), false)
        );
    }

    private static List<Key> materials(Object current) {
        return List.of(
                Key.of( "gedruckte Ressource", "text", "text".equals(current), false),
                Key.of( "Online-Ressource",  "online_resource", "online_resource".equals(current), false),
                Key.of( "elektronische Ressource",  "electronic_resource", "electronic_resource".equals(current), false),
                Key.of( "Mikroform",  "microform", "microform".equals(current), false),
                Key.of( "Dia",  "projected_graphic", "projected_graphic".equals(current), false),
                Key.of( "Bildvorlage",  "non_projected_graphic", "non_projected_graphic".equals(current), false),
                Key.of( "bewegte Bilder",  "motion_picture", "motion_picture".equals(current), false),
                Key.of( "Blindenschrift",  "tactile", "tactile".equals(current), false),
                Key.of( "Notenschrift",  "notated_music", "notated_music".equals(current), false),
                Key.of( "Satellitenbilder",  "remote_sensing_image", "remote_sensing_image".equals(current), false),
                Key.of( "Tonaufnahme",  "sound_recording", "sound_recording".equals(current), false),
                Key.of( "Bildaufnahme",  "video_recording", "video_recording".equals(current), false),
                Key.of( "Bausatz",  "kit", "kit".equals(current), false),
                Key.of( "Landkarte",  "map", "map".equals(current), false),
                Key.of( "Globus",  "globe", "globe".equals(current), false),
                Key.of( "nicht spezifiziert", "unspecified", "unspecified".equals(current), false),
                Key.of( "nicht definiert", "", "".equals(current), false)
        );
    }

    private static List<Key> domains(Object current) {
        return List.of(
                Key.of( "BAW", "BAW", "BAW".equals(current), false),
                Key.of( "BAY", "BAY", "BAY".equals(current), false),
                Key.of( "BER", "BER", "BER".equals(current), false),
                Key.of( "GBV", "GBV", "GBV".equals(current), false),
                Key.of( "HAM", "HAM", "HAM".equals(current), false),
                Key.of( "HES", "HES", "HES".equals(current), false),
                Key.of( "NIE", "NIE", "NIE".equals(current), false),
                Key.of( "NRW", "NRW", "NRW".equals(current), false),
                Key.of( "SAA", "SAA", "SAA".equals(current), false),
                Key.of( "SAX", "SAX", "SAX".equals(current), false),
                Key.of( "THU", "THU", "THU".equals(current), false),
                Key.of( "nicht definiert", "", "".equals(current), false)
        );
    }

    public static Collection<?> getList(Map<String, Object> map, String key) {
        Object obj = map.get(key);
        if (obj instanceof Collection<?> collection) {
            return collection;
        }
        return obj != null ? List.of(obj) : List.of();
    }

    /**
     * Return the most current value given by a key. If the value is not present as a form value, try looking up the path.
     * @param key the key
     * @param parameter the parameter
     * @param map the backing map (persistent in a session)
     * @param defaultValue the default value if there is no other value
     * @return the current value
     */
    private static Object current(String key, Parameter parameter, Map<String, Object> map, Object defaultValue) {
        boolean b1 = parameter.containsKey(key, Parameter.Domain.FORM);
        boolean b2 = parameter.containsKey(key, Parameter.Domain.PATH);
        Object defaultObject = map != null ? map.getOrDefault(key, defaultValue) : defaultValue;
        Object object = defaultObject;
        try {
            if (b2) {
                object = asSingleValueIfPossible(parameter.getAll(key, Parameter.Domain.PATH));
            }
            if (b1) {
                object = asSingleValueIfPossible(parameter.getAll(key, Parameter.Domain.FORM));
            }
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        logger.log(Level.FINEST, "current: key = " + key + " value = '" + object + "' in form = " + b1 +
                " in path = " + b2 +
                " default = " + defaultObject);
        return object;
    }
    /**
     * Checkboxed values. This is driven by HTTP/HTML oddities.
     * If no checkbox is checked, the HTTP parameter is not sent at all.
     * That means we have a different semantic, and we must detect if there is an HTML form POST request present at all.
     * If such a form is present and no parameter is passed, reset to an empty list.
     * But if no form is present, retrieve value from the persistent map.
     * This is useful for handling "post"/"see other"/"get" chain, where the last GET request sends no parameters.
     * @param key the key
     * @param parameter the parameter
     * @return list of values for checkbox representation
     */
    private static Object currentCheck(String key, Parameter parameter, Map<String, Object> map, Object defaultValue) {
        if (parameter.isPresent(Parameter.Domain.FORM)) {
            try {
                return parameter.containsKey(key, Parameter.Domain.FORM) ?
                        asSingleValueIfPossible(parameter.getAll(key, Parameter.Domain.FORM)) : defaultValue;
            } catch (Exception e) {
                // can't happen
                logger.log(Level.SEVERE, "can't happen");
                return null;
            }
        } else {
            return current(key, parameter, map, defaultValue);
        }
    }

    /**
     * Unwraps a collection to the first object if the collection is of size 1, otherwise do not unwrap.
     * @param obj if collection of size 1, either the object in the collection, or the collection
     * @return the first object
     */
    private static Object asSingleValueIfPossible(Object obj) {
        if (obj instanceof Collection<?> collection) {
            if (collection.size() == 1) {
                return collection.iterator().next();
            }
        }
        return obj;
    }

    @SuppressWarnings("serial")
    public static class Key extends LinkedHashMap<String, String> {

        private final String label;

        private final String value;

        private final boolean isSelected;

        private final boolean isDisabled;

        private Key(String label, String value, boolean isSelected, boolean isDisabled) {
            super();
            this.label = label;
            this.value = value;
            this.isSelected = isSelected;
            this.isDisabled = isDisabled;
            put("label", label);
            put("value", value);
            put("is_selected", String.valueOf(isSelected));
            put("is_disabled", String.valueOf(isDisabled));
        }

        public static Key of(String label, String value, boolean isSelected, boolean isDisabled) {
            return new Key(label, value, isSelected, isDisabled);
        }

        public String getLabel() {
            return label;
        }

        public String getValue() {
            return value;
        }

        public boolean isSelected() {
            return isSelected;
        }

        public boolean isDisabled() {
            return isDisabled;
        }
    }
}
