package org.xbib.interlibrary.http.server.avail;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.function.Predicate;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xbib.datastructures.common.MultiMap;
import org.xbib.interlibrary.action.avail.AvailAction;
import org.xbib.interlibrary.action.avail.AvailRequest;
import org.xbib.interlibrary.action.avail.AvailRequester;
import org.xbib.interlibrary.action.avail.AvailResponse;
import org.xbib.interlibrary.action.avail.DefaultAvailRequester;
import org.xbib.interlibrary.action.resolve.DefaultResolveRequester;
import org.xbib.interlibrary.action.resolve.ResolveAction;
import org.xbib.interlibrary.action.resolve.ResolveRequest;
import org.xbib.interlibrary.action.resolve.ResolveRequester;
import org.xbib.interlibrary.action.resolve.ResolveResponse;
import org.xbib.interlibrary.api.ApplicationModule;
import org.xbib.interlibrary.api.library.Library;
import org.xbib.interlibrary.api.library.LibraryService;
import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.api.library.Mode;
import org.xbib.interlibrary.common.DefaultBibliographicDescription;
import org.xbib.interlibrary.common.Interlibrary;
import org.xbib.interlibrary.common.InterlibraryModule;
import org.xbib.interlibrary.common.predicate.MaterialPredicate;
import org.xbib.interlibrary.common.predicate.DomainPredicate;
import org.xbib.interlibrary.common.predicate.LastResortPredicate;
import org.xbib.interlibrary.common.predicate.LibraryGroupPredicate;
import org.xbib.interlibrary.common.predicate.LibraryPredicate;
import org.xbib.interlibrary.common.predicate.ModePredicate;
import org.xbib.interlibrary.common.predicate.OrganizationPredicate;

public class AvailModule implements InterlibraryModule {

    private static final Logger logger = Logger.getLogger(AvailModule.class.getName());

    private Interlibrary interlibrary;

    public AvailModule() {
    }

    @Override
    public String getName() {
        return "AvailModule";
    }

    @Override
    public List<Class<? extends ApplicationModule>> dependsOn() {
        return null;
    }

    @Override
    public void setInterlibrary(Interlibrary interlibrary) {
        this.interlibrary = interlibrary;
    }

    @Override
    public Interlibrary getInterlibrary() {
        return interlibrary;
    }

    public AvailResponse avail(MultiMap<String, Object> params) {
        if (params == null) {
            return null;
        }
        String requesterISIL = params.getString(AvailParameters.REQUESTER, "");
        Library requesterLibrary = getInterlibrary().getConfiguration().getLibrary(requesterISIL);
        if (requesterLibrary == null) {
            throw new IllegalArgumentException("requester library not found for " + requesterISIL);
        }
        AvailRequester requester = new DefaultAvailRequester(requesterISIL,
                params.getString(AvailParameters.USERGROUP, ""),
                params.getBoolean(AvailParameters.SELF_ALLOWED, false),
                params.getBoolean(AvailParameters.LOCAL_ALLOWED, false),
                params.getBoolean(AvailParameters.DOMAIN_ALLOWED, true), // default mode is to use domains
                params.getBoolean(AvailParameters.LAST_RESORT_ALLOWED, false));
        DefaultBibliographicDescription.Builder bibliographicDescriptionBuilder = DefaultBibliographicDescription.builder()
                .setSource(params.getString(AvailParameters.SOURCE, ""))
                .setSourceId(params.getString(AvailParameters.SOURCE_ID,""));
        String year = params.getString(AvailParameters.YEAR, null);
        if (year != null && !year.isEmpty()) {
            // clean input to obtain a four-digit gregorian year
            bibliographicDescriptionBuilder.setYear(Integer.parseInt(cleanYear(year)));
        }
        if (params.containsKey(AvailParameters.ISSN)) {
            bibliographicDescriptionBuilder.addIdentifier(Map.of("type", "ISSN", "value", params.getString(AvailParameters.ISSN)));
        }
        if (params.containsKey(AvailParameters.ISBN)) {
            bibliographicDescriptionBuilder.addIdentifier(Map.of("type", "ISBN", "value", params.getString(AvailParameters.ISBN)));
        }
        if (params.containsKey(AvailParameters.DOI)) {
            bibliographicDescriptionBuilder.addIdentifier(Map.of("type", "DOI", "value", params.getString(AvailParameters.DOI)));
        }
        Collection<?> advantageParam = params.get(AvailParameters.ADVANTAGE_LIBRARY);
        List<String> advantage = advantageParam != null ? advantageParam.stream().map(Object::toString).toList() : null;
        Collection<?> handicapParam = params.get(AvailParameters.HANDICAP_LIBRARY);
        List<String> handicap = handicapParam != null ? handicapParam.stream().map(Object::toString).toList() : null;
        AvailRequest.Builder builder = AvailRequest.builder()
                .setRequester(requester)
                .setBibliographicDescription(bibliographicDescriptionBuilder.build())
                .setBibliographicDescriptionEnabled(params.getBoolean(AvailParameters.ENABLE_DESCRIPTION, false))
                .setLinkPriorityEnabled(params.getBoolean(AvailParameters.ENABLE_LINK_PRIORITY, true))
                .setServices(params.get(AvailParameters.SERVICE))
                .setDomainEnabled(params.getBoolean(AvailParameters.ENABLE_DOMAIN, true)) // default mode is to use domain
                .setDebugEnabled(params.getBoolean(AvailParameters.DEBUG, false))
                .setOnlineResourceEnabled(params.getBoolean(AvailParameters.ENABLE_ONLINE_RESOURCES, true))
                .setPrintResourceEnabled(params.getBoolean(AvailParameters.ENABLE_PRINT_RESOURCES, true))
                .setAdvantage(advantage)
                .setHandicap(handicap);
        Predicate<Library> libraryPredicate = library -> true;
        StringBuilder libraryPredicateText = new StringBuilder();
        Predicate<LibraryService> libraryServicePredicate = libraryService -> true;
        StringBuilder libraryServicePredicateText = new StringBuilder();
        if (params.containsKey(AvailParameters.INCLUDE_DOMAIN)) {
            Predicate<Library> domainPredicate = null;
            StringBuilder domainPredicateText = new StringBuilder();
            for (Object o : params.get(AvailParameters.INCLUDE_DOMAIN)) {
                String domain = o.toString();
                // if domain parameter is an empty string, skip it
                if (!"".equals(domain)) {
                    domainPredicate = domainPredicate != null ?
                            domainPredicate.or(new DomainPredicate(domain)) : new DomainPredicate(domain);
                    domainPredicateText.append(" OR DOMAIN(").append(domain).append(")");
                }
            }
            if (domainPredicate != null) {
                libraryPredicate = libraryPredicate.and(domainPredicate);
                libraryPredicateText.append(" AND (").append(domainPredicateText).append(")");
            }
        }
        if (params.containsKey(AvailParameters.EXCLUDE_DOMAIN)) {
            Predicate<Library> domainPredicate = null;
            StringBuilder domainPredicateText = new StringBuilder();
            for (Object o : params.get(AvailParameters.EXCLUDE_DOMAIN)) {
                String domain = o.toString();
                if (!"".equals(domain)) {
                    domainPredicate = domainPredicate != null ?
                            domainPredicate.and(new DomainPredicate(domain).negate()) : new DomainPredicate(domain).negate();
                    domainPredicateText.append(" AND NOT(DOMAIN(").append(domain).append(")");
                }
            }
            if (domainPredicate != null) {
                libraryPredicate = libraryPredicate.and(domainPredicate);
                libraryPredicateText.append(" AND (").append(domainPredicateText).append(")");
            }
        }
        if (params.containsKey(AvailParameters.INCLUDE_ORGANIZATION)) {
            Predicate<Library> organizationPredicate = null;
            StringBuilder organizationPredicateText = new StringBuilder();
            for (Object o : params.get(AvailParameters.INCLUDE_ORGANIZATION)) {
                String organization = o.toString();
                // if organization parameter is an empty string, skip it
                if (!"".equals(organization)) {
                    organizationPredicate = organizationPredicate != null ?
                            organizationPredicate.or(new OrganizationPredicate(organization)) : new OrganizationPredicate(organization);
                    organizationPredicateText.append(" OR ORGANIZATION(").append(organization).append(")");
                }
            }
            if (organizationPredicate != null) {
                libraryPredicate = libraryPredicate.and(organizationPredicate);
                libraryPredicateText.append(" AND (").append(organizationPredicateText).append(")");
            }
        }
        if (params.containsKey(AvailParameters.EXCLUDE_ORGANIZATION)) {
            Predicate<Library> organizationPredicate = null;
            StringBuilder organizationPredicateText = new StringBuilder();
            for (Object o : params.get(AvailParameters.EXCLUDE_ORGANIZATION)) {
                String organization = o.toString();
                if (!"".equals(organization)) {
                    organizationPredicate = organizationPredicate != null ?
                            organizationPredicate.and(new OrganizationPredicate(organization).negate()) : new OrganizationPredicate(organization).negate();
                    organizationPredicateText.append(" AND NOT(ORGANIZATION(").append(organization).append(")");
                }
            }
            if (organizationPredicate != null) {
                libraryPredicate = libraryPredicate.and(organizationPredicate);
                libraryPredicateText.append(" AND (").append(organizationPredicateText).append(")");
            }
        }
        if (params.containsKey(AvailParameters.LIBRARY_GROUP)) {
            Predicate<Library> groupPredicate = null;
            StringBuilder groupPredicateText = new StringBuilder();
            for (Object o : params.get(AvailParameters.LIBRARY_GROUP)) {
                String group = o.toString();
                if (!"".equals(group)) {
                    groupPredicate = groupPredicate != null ?
                            groupPredicate.or(new LibraryGroupPredicate(group)) : new LibraryGroupPredicate(group);
                    groupPredicateText.append(" OR GROUP(").append(group).append(")");
                }
            }
            if (groupPredicate != null) {
                libraryPredicate = libraryPredicate.and(groupPredicate);
                libraryPredicateText.append(" AND (").append(groupPredicateText).append(")");
            }
        }
        if (params.containsKey(AvailParameters.INCLUDE_LIBRARY)) {
            Predicate<Library> libraryIncludePredicate = null;
            StringBuilder libraryIncludePredicateText = new StringBuilder();
            for (Object o : params.get(AvailParameters.INCLUDE_LIBRARY)) {
                String library = o.toString();
                if (!"".equals(library)) {
                    libraryIncludePredicate = libraryIncludePredicate != null ?
                            libraryIncludePredicate.or(new LibraryPredicate(library)) : new LibraryPredicate(library);
                    libraryIncludePredicateText.append(" OR LIBRARY(").append(library).append(")");
                }
            }
            if (libraryIncludePredicate != null) {
                libraryPredicate = libraryPredicate.and(libraryIncludePredicate);
                libraryPredicateText.append(" AND (").append(libraryIncludePredicateText).append(")");
            }
        }
        if (params.containsKey(AvailParameters.EXCLUDE_LIBRARY)) {
            Predicate<Library> libraryExcludePredicate = null;
            StringBuilder libraryExcludePredicateText = new StringBuilder();
            for (Object o : params.get(AvailParameters.EXCLUDE_LIBRARY)) {
                String library = o.toString();
                if (!"".equals(library)) {
                    libraryExcludePredicate = libraryExcludePredicate != null ?
                            libraryExcludePredicate.and(new LibraryPredicate(library).negate()) : new LibraryPredicate(library).negate();
                    libraryExcludePredicateText.append(" AND NOT(LIBRARY(").append(library).append("))");
                }
            }
            if (libraryExcludePredicate != null) {
                libraryPredicate = libraryPredicate.and(libraryExcludePredicate);
                libraryPredicateText.append(" AND (").append(libraryExcludePredicateText).append(")");
            }
        }
        List<Mode> modes = new ArrayList<>();
        if (params.containsKey(AvailParameters.MODE)) {
            Predicate<LibraryService> modePredicate = null;
            StringBuilder modePredicateText = new StringBuilder();
            for (Object o : params.get(AvailParameters.MODE)) {
                String modeString = o.toString();
                if (!"".equals(modeString)) {
                    Mode mode = Mode.fromValue(modeString);
                    modes.add(mode);
                    modePredicate = modePredicate != null ?
                            modePredicate.or(new ModePredicate(mode)) : new ModePredicate(mode);
                    modePredicateText.append(" OR MODE(").append(mode.toValue()).append(")");
                }
            }
            if (modePredicate != null) {
                libraryServicePredicate = libraryServicePredicate.and(modePredicate);
                libraryServicePredicateText.append(" AND(").append(modePredicateText).append(")");
            }
        }
        Predicate<Library> allowZFL =
                new LibraryGroupPredicate("ZFL").or(new DomainPredicate("NRW").negate());
        Predicate<Library> notblockedZFL =
                new LibraryGroupPredicate("ZFLblock").negate();
        Predicate<Library> allowSFL =
                new LibraryGroupPredicate("SFL").or(new DomainPredicate("NRW").negate());
        Predicate<Library> notblockedSFL =
                new LibraryGroupPredicate("SFLblock").negate();
        Predicate<Library> allowEFL =
                new LibraryGroupPredicate("EFL").or(new DomainPredicate("NRW").negate());
        Predicate<Library> notblockedEFL =
                new LibraryGroupPredicate("EFLblock").negate();
        Material material = getMaterial(params);
        switch (material) {
            case ONLINE_RESOURCE -> {
                logger.log(Level.FINE, "special filtering for material " + material);
                libraryServicePredicate = libraryServicePredicate.and(new MaterialPredicate(Material.ONLINE_RESOURCE));
                libraryServicePredicateText.append(" AND MATERIAL(").append(Material.ONLINE_RESOURCE.toValue()).append(")");
                builder.setOnlineResourceEnabled(true);
                builder.setPrintResourceEnabled(false);
                if (modes.contains(Mode.COPY)) {
                    libraryPredicate = libraryPredicate.and(allowEFL.and(notblockedEFL));
                    libraryPredicateText.append(" AND (GROUP(EFL) OR NOT(DOMAIN(NRW)) AND NOT(GROUP(EFLblock)))");
                } else if (modes.contains(Mode.LOAN)) {
                    libraryPredicate = libraryPredicate.and(allowZFL.and(notblockedZFL));
                    libraryPredicateText.append(" AND (GROUP(ZFL) OR NOT(DOMAIN(NRW)) AND NOT(GROUP(ZFLblock)))");
                }
            }
            case TEXT -> {
                logger.log(Level.FINE, "special filtering for material " + material);
                libraryServicePredicate = libraryServicePredicate.and(new MaterialPredicate(Material.TEXT));
                builder.setOnlineResourceEnabled(false);
                builder.setPrintResourceEnabled(true);
                if (modes.contains(Mode.COPY)) {
                    libraryPredicate = libraryPredicate.and(allowSFL.and(notblockedSFL));
                    libraryPredicateText.append(" AND (GROUP(SFL) OR NOT(DOMAIN(NRW)) AND NOT(GROUP(SFLblock)))");
                } else if (modes.contains(Mode.LOAN)) {
                    libraryPredicate = libraryPredicate.and(allowZFL.and(notblockedZFL));
                    libraryPredicateText.append(" AND (GROUP(ZFL) OR NOT(DOMAIN(NRW)) AND NOT(GROUP(ZFLblock)))");
                }
            }
            default -> {
                logger.log(Level.FINE, "material " + material + " ignored");
                if (modes.contains(Mode.COPY)) {
                    libraryPredicate = libraryPredicate.and(allowSFL.or(allowEFL)).and(notblockedSFL.and(notblockedEFL));
                    libraryPredicateText.append(" AND (GROUP(SFL) OR GROUP(EFL) OR NOT(DOMAIN(NRW)) AND NOT(GROUP(SFLblock)) AND NOT(GROUP(EFLblock)))");
                } else if (modes.contains(Mode.LOAN)) {
                    libraryPredicate = libraryPredicate.and(allowZFL.and(notblockedZFL));
                    libraryPredicateText.append(" AND (GROUP(ZFL) OR NOT(DOMAIN(NRW)) AND NOT(GROUP(ZFLblock)))");
                }
            }
        }
        libraryPredicate = libraryPredicate.or(new LastResortPredicate());
        libraryPredicateText.append(" OR LASTRESORT()");
        builder.setLibraryPredicate(libraryPredicate)
                .setLibraryPredicateText(libraryPredicateText.toString())
                .setLibraryServicePredicate(libraryServicePredicate)
                .setLibraryServicePredicateText(libraryServicePredicateText.toString())
                .setSize(params.getInteger(AvailParameters.LIMIT, 5))
                .setFilteredLibrariesEnabled(params.getBoolean(AvailParameters.ENABLE_FILTERED, false));
        return interlibrary.execute(AvailAction.INSTANCE, builder.build());
    }

    public ResolveResponse resolve(MultiMap<String, Object> params) {
        ResolveRequest.Builder builder = ResolveRequest.builder();
        if (params.containsKey(ResolveParameters.SERVICE)) {
            for (Object service : params.get(ResolveParameters.SERVICE)) {
                // if service is en empty string, skip it
                if (!"".equals(service)) {
                    builder.setServices(service.toString());
                }
            }
        }
        ResolveRequester requester = new DefaultResolveRequester(params.getString(ResolveParameters.REQUESTER, ""));
        builder.setRequester(requester);
        String year = params.getString(ResolveParameters.YEAR, null);
        DefaultBibliographicDescription.Builder descriptionBuilder = DefaultBibliographicDescription.builder()
                .setSource(params.getString(ResolveParameters.SOURCE))
                .setSourceId(params.getString(ResolveParameters.SOURCE_ID));
        if (year != null && !year.isEmpty()) {
            descriptionBuilder.setYear(Integer.parseInt(year));
        }
        if (params.containsKey(AvailParameters.ISSN)) {
            descriptionBuilder.addIdentifier(Map.of("type", "ISSN", "value", params.getString(AvailParameters.ISSN)));
        }
        if (params.containsKey(AvailParameters.ISBN)) {
            descriptionBuilder.addIdentifier(Map.of("type", "ISBN", "value", params.getString(AvailParameters.ISBN)));
        }
        if (params.containsKey(AvailParameters.DOI)) {
            descriptionBuilder.addIdentifier(Map.of("type", "DOI", "value", params.getString(AvailParameters.DOI)));
        }
        builder.setBibliographicDescription(descriptionBuilder.build());
        builder.setOnlineResourceEnabled(params.getBoolean(ResolveParameters.ENABLE_ONLINE_RESOURCE, false));
        return interlibrary.execute(ResolveAction.INSTANCE, builder.build());
    }

    @Override
    public void run() {
    }

    @Override
    public void close() throws IOException {
    }

    /**
     * Clean input to obtain a digit year with at most four decimal digits.
     * @param string the dirty input string
     * @return at maximum four decimal digits
     */
    private static String cleanYear(String string) {
        if (string == null) {
            return null;
        }
        String clean = string;
        clean = clean.replaceAll("[^\\d]", "")
                .trim();
        if (clean.length() > 4) {
            clean = clean.substring(0, 4);
        }
        return clean;
    }

    private static Material getMaterial(MultiMap<String, Object> multiMap) {
        if (multiMap.containsKey(AvailParameters.MATERIAL)) {
            Collection<Object> materials = multiMap.get(AvailParameters.MATERIAL);
            String materialString = materials.isEmpty() ? "UNSPECIFIED" : materials.iterator().next().toString();
            return Material.fromValue(materialString);
        }
        return Material.UNSPECIFIED;
    }
}
