package org.xbib.interlibrary.http.server.avail;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Objects;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.json.tiny.JsonBuilder;
import org.xbib.interlibrary.action.avail.AvailAction;
import org.xbib.interlibrary.action.avail.AvailResponse;
import org.xbib.interlibrary.action.avail.AvailService;
import org.xbib.interlibrary.action.resolve.ResolveResponse;
import org.xbib.interlibrary.api.action.Action;
import org.xbib.interlibrary.api.action.Service;
import org.xbib.interlibrary.http.server.InterlibraryServer;
import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.http.HttpAddress;
import org.xbib.net.http.HttpHeaderNames;
import org.xbib.net.http.HttpHeaderValues;
import org.xbib.net.http.HttpMethod;
import org.xbib.net.http.HttpResponseStatus;
import org.xbib.net.http.j2html.J2HtmlService;
import org.xbib.net.http.server.application.Application;
import org.xbib.net.http.server.domain.BaseHttpDomainBuilder;
import org.xbib.net.http.server.domain.HttpDomain;
import org.xbib.net.http.server.domain.HttpSecurityDomain;
import org.xbib.net.http.server.resource.ClassLoaderResourceHandler;
import org.xbib.net.http.server.service.BaseHttpService;

public final class HttpAvailServer extends InterlibraryServer {

    private static final Logger logger = Logger.getLogger(HttpAvailServer.class.getName());

    public HttpAvailServer() {
        super();
    }

    @Override
    public HttpDomain buildDomain(HttpAddress httpAddress,
                                  HttpSecurityDomain securityDomain,
                                  BaseHttpDomainBuilder domainBuilder) {
        AvailModule availModule = Objects.requireNonNull(getModule(AvailModule.class));
        String prefix = Objects.requireNonNull(getSettings().get("web.prefix"), "web.prefix must not be null");
        AvailFormAction availFormAction = new AvailFormAction();
        AvailFormViewHandler availFormViewHandler = new AvailFormViewHandler();
        return domainBuilder
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/favicon.ico")
                    .setHandler(ctx -> {
                        ctx.status(HttpResponseStatus.OK)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, "image/x-icon")
                                .body(fromHex(hexFavIcon))
                                .done();
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/resources/dist/**")
                    .setHandler(new ClassLoaderResourceHandler(HttpAvailServer.class.getClassLoader(), "META-INF/", 24 * 2600))
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/resources/webjars/**")
                    .setHandler(new ClassLoaderResourceHandler(HttpAvailServer.class.getClassLoader(), "META-INF/", 24 * 3600))
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/status")
                    .setHandler(ctx -> {
                        ctx.status(HttpResponseStatus.OK)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.TEXT_PLAIN)
                                .charset(StandardCharsets.UTF_8)
                                .body(getActionStates());
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/service/{service}/{status}")
                    .setMethod(HttpMethod.PUT)
                    .setHandler(ctx -> {
                        Parameter parameter = ctx.getRequest().getParameter();
                        try {
                            Object serviceName = parameter.get("service", Parameter.Domain.PATH);
                            Object serviceStatus = parameter.get("status", Parameter.Domain.PATH);
                            if (serviceName != null && serviceStatus != null) {
                                if ("true".equals(serviceStatus) || "false".equals(serviceStatus)) {
                                    setActionService(AvailAction.INSTANCE, serviceName.toString(), Boolean.parseBoolean(serviceStatus.toString()));
                                }
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.TEXT_PLAIN)
                                        .charset(StandardCharsets.UTF_8)
                                        .body("OK");
                            } else {
                                ctx.status(HttpResponseStatus.BAD_REQUEST);
                            }
                        } catch (ParameterException e) {
                            logger.log(Level.WARNING, e.getMessage(), e);
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/resolve/api/{requester}/{source}/{sourceid}/{year}")
                    .setHandler(ctx -> {
                        ResolveResponse resolveResponse = availModule.resolve(ctx.getRequest().asMultiMap());
                        logger.log(Level.INFO, "stat: request=" + resolveResponse.getRequest() +
                                " status=" + resolveResponse.getStatus() +
                                " service=" + resolveResponse.getService() +
                                " bibliographicdescription=" + resolveResponse.getBibliographicDescription()
                        );
                        HttpResponseStatus status = switch (resolveResponse.getStatus()) {
                            case NOT_FOUND -> HttpResponseStatus.NOT_FOUND;
                            case OK, EMPTY_RESPONSE -> HttpResponseStatus.OK;
                            default -> HttpResponseStatus.INTERNAL_SERVER_ERROR;
                        };
                        ctx.status(status)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                .charset(StandardCharsets.UTF_8)
                                .body(resolveResponse.asJson())
                                .done();
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/resolve/api/{requester}/{source}/{sourceid}")
                    .setHandler(ctx -> {
                        ResolveResponse resolveResponse = availModule.resolve(ctx.getRequest().asMultiMap());
                        logger.log(Level.INFO, "stat: request=" + resolveResponse.getRequest() +
                                " status=" + resolveResponse.getStatus() +
                                " service=" + resolveResponse.getService() +
                                " bibliographicdescription=" + resolveResponse.getBibliographicDescription()
                        );
                        HttpResponseStatus status = switch (resolveResponse.getStatus()) {
                            case NOT_FOUND -> HttpResponseStatus.NOT_FOUND;
                            case OK -> HttpResponseStatus.OK;
                            case EMPTY_RESPONSE -> HttpResponseStatus.SERVICE_UNAVAILABLE;
                            default -> HttpResponseStatus.INTERNAL_SERVER_ERROR;
                        };
                        ctx.status(status)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                .charset(StandardCharsets.UTF_8)
                                .body(resolveResponse.asJson())
                                .done();
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/resolve/api")
                    .setHandler(ctx -> {
                        ResolveResponse resolveResponse = availModule.resolve(ctx.getRequest().asMultiMap());
                        logger.log(Level.INFO, "stat: request=" + resolveResponse.getRequest() +
                                " status=" + resolveResponse.getStatus() +
                                " service=" + resolveResponse.getService() +
                                " bibliographicdescription=" + resolveResponse.getBibliographicDescription()
                        );
                        HttpResponseStatus status = switch (resolveResponse.getStatus()) {
                            case NOT_FOUND -> HttpResponseStatus.NOT_FOUND;
                            case OK, EMPTY_RESPONSE -> HttpResponseStatus.OK;
                            default -> HttpResponseStatus.INTERNAL_SERVER_ERROR;
                        };
                        ctx.status(status)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                .charset(StandardCharsets.UTF_8)
                                .body(resolveResponse.asJson())
                                .done();
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/api/{requester}/{source}/{sourceid}/{year}")
                    .setHandler(ctx -> {
                        AvailResponse availResponse = availModule.avail(ctx.getRequest().asMultiMap());
                        availResponse.log();
                        HttpResponseStatus status = switch (availResponse.getStatus()) {
                            case NOT_FOUND -> HttpResponseStatus.NOT_FOUND;
                            case OK, EMPTY_RESPONSE -> HttpResponseStatus.OK;
                            default -> HttpResponseStatus.INTERNAL_SERVER_ERROR;
                        };
                        ctx.status(status)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                .charset(StandardCharsets.UTF_8)
                                .body(availResponse.asJson())
                                .done();
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/api/{requester}/{source}/{sourceid}")
                    .setHandler(ctx -> {
                        AvailResponse availResponse = availModule.avail(ctx.getRequest().asMultiMap());
                        availResponse.log();
                        HttpResponseStatus status = switch (availResponse.getStatus()) {
                            case NOT_FOUND -> HttpResponseStatus.NOT_FOUND;
                            case OK, EMPTY_RESPONSE -> HttpResponseStatus.OK;
                            default -> HttpResponseStatus.INTERNAL_SERVER_ERROR;
                        };
                        ctx.status(status)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                .charset(StandardCharsets.UTF_8)
                                .body(availResponse.asJson())
                                .done();
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/api")
                    .setHandler(ctx -> {
                        AvailResponse availResponse = availModule.avail(ctx.getRequest().asMultiMap());
                        availResponse.log();
                        HttpResponseStatus status = switch (availResponse.getStatus()) {
                            case NOT_FOUND -> HttpResponseStatus.NOT_FOUND;
                            case OK, EMPTY_RESPONSE -> HttpResponseStatus.OK;
                            default -> HttpResponseStatus.INTERNAL_SERVER_ERROR;
                        };
                        ctx.status(status)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                .charset(StandardCharsets.UTF_8)
                                .body(availResponse.asJson())
                                .done();
                    })
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/action/availform/{availform}")
                    .setMethod(HttpMethod.POST, HttpMethod.GET)
                    .setHandler(availFormAction)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/form/{availform}")
                    .setMethod(HttpMethod.POST, HttpMethod.GET)
                    .setHandler(availFormViewHandler)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/form")
                    .setMethod(HttpMethod.POST, HttpMethod.GET)
                    .setHandler(availFormViewHandler)
                    .build())
            .build();
    }

    @Override
    protected void customizeApplication(Application application) {
        application.addModule(new AvailApplicationModule(this, application, "app", getSettings()));
    }

    private String getActionStates() {
        StringBuilder sb = new StringBuilder();
        sb.append("Actions\n-------\n");
        for (Action<?, ?, ?, ?> action : getActions()) {
            sb.append(action.getInstance().getClass().getName()).append("\n");
            for (Service<?, ?, ?> service : action.getInstance().getServices()) {
                sb.append("  ").append(service.getName())
                        .append("  ").append(service.isEnabled());
                if (service.getLastThrowable() != null) {
                    sb.append("  ").append(service.getLastThrowable().getMessage());
                }
                sb.append("\n");
            }
        }
        return sb.toString();
    }

    private String getAvailServicesAsJsonString() throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.beginMap();
        builder.beginCollection("service");
        for (Action<?, ?, ?, ?> action : getActions()) {
            if (action instanceof AvailAction availAction) {
                for (AvailService availService : availAction.getServices()) {
                    if (availService.isEnabled()) {
                        builder.beginMap();
                        builder.field("name", availService.getName());
                        builder.field("label", availService.getLabel().toUpperCase(Locale.ROOT));
                        builder.field("sort", availService.getSortKey());
                        builder.field("url", availService.getURL("%s"));
                        builder.field("is_disabled", false);
                        builder.endMap();
                    }
                }
            }
        }
        builder.endCollection();
        builder.endMap();
        return builder.build();
    }

    public List<AvailForm.Key> getAvailServices(Collection<?> currentValue) {
        List<AvailForm.Key> list = new LinkedList<>();
        for (Action<?, ?, ?, ?> action : getActions()) {
            if (action instanceof AvailAction availAction) {
                for (AvailService availService : availAction.getServices()) {
                    if (availService.isEnabled()) {
                        AvailForm.Key key = AvailForm.Key.of(availService.getLabel(), availService.getName(),
                                currentValue != null && currentValue.contains(availService.getName()), false);
                        list.add(key);
                    }
                }
                // always add ISBN, ISSN, DOI
                for (String source : new String[] { "ISBN", "ISSN", "DOI"}) {
                    AvailForm.Key key = AvailForm.Key.of(source, source,
                            currentValue != null && currentValue.contains(source), false);
                    list.add(key);
                }
            }
        }
        return list;
    }
}
